"""Solvers.

This submodule contains the solvers for the computation of thermal processes in
2D.

"""

from .explicit_k import explicit_k
from .explicit_general import explicit_general

__all__ = [explicit_k, explicit_general]



"""Objects.

This submodule includes the object class for creating thermal objects,
object system class for creating systems of thermal objects, and
a complex single object for using in more complex thermal computing.

"""

from .object import Object
from .system import SystemObjects
from .single import SingleObject

__all__ = [Object, SystemObjects, SingleObject]



"""Heatrapy.

This module allows to create thermal objects, establish thermal contact between
them, activate or deactivate the whole, or part, of the materials, and compute
the respective heat transfer processes, in 1D and 2D. It includes two system
models for the computation of caloric systems in 1D.

The package is based on three classes that create general models:

######
Object1D
######

This class only creates a single unidimensional thermal object. It includes two
methods: material activation and material deactivation, of part of the object.

######
Object2D
######

This class only creates a single two-dimensional thermal object. It includes
two methods: material activation and material deactivation, of part of the
object.

##############
SystemObjects1D
##############

This class creates a system of unidimensional thermal objects that can be in
contact to each other and computes the respective heat transfer processes. It
uses the class Object1D for the creation of each thermal object.

##############
SystemObjects2D
##############

This class creates a system of two-dimensional thermal objects that can be in
contact to each other and computes the respective heat transfer processes. It
uses the class Object2D for the creation of each thermal object.

#############
SingleObject1D
#############

This class computes the heat transfer processes involved in only one
unidimensional thermal object. It uses the class Object1D for activating and
deactivating the material.

#############
SingleObject2D
#############

This class computes the heat transfer processes involved in only one
two-dimensional thermal object. It uses the class Object1D for activating and
deactivating the material.


At this moment there are two type of caloric systems (functions) that can be
computed:

########################
fluid_active_regenerator
########################

This function creates and computes a unidimensional active regenerative system
used for refrigeration and heat pumps. The heat exchanger is a fluid. It can be
used to compute caloric systems, e.g. magnetocaloric, electrocaloric,
elastocaloric, and barocaloric.

########################
solid_active_regenerator
########################

This function creates and computes a unidimensional active regenerative system
used for refrigeration and heat pumps. The heat exchanger is the solid material
itself. It can be used to compute caloric systems, e.g. magnetocaloric,
electrocaloric, elastocaloric, and barocaloric.

"""

from .dimension_1.objects import Object as Object1D
from .dimension_1.objects import SystemObjects as SystemObjects1D
from .dimension_1.objects import SingleObject as SingleObject1D
from .dimension_2.objects import Object as Object2D
from .dimension_2.objects import SystemObjects as SystemObjects2D
from .dimension_2.objects import SingleObject as SingleObject2D

__all__ = [Object1D, SystemObjects1D, SingleObject1D, Object2D,
           SystemObjects2D, SingleObject2D]



"""Contains the classe single_object.

Used to compute system models

"""

from .. import solvers
from . import Object
import matplotlib.pyplot as plt
import numpy as np


class SingleObject:
    """Single_object class.

    This class solves numerically the heat conduction equation for 1 dimension
    of an active material(s). Three solvers can be used: explicit with
    x-independent k, explicit with x-dependent k, implicit with x-independent
    k, and implicit with x-dependent k. The class has 5 methods: activate for
    the activation of part of the solid, deactivate for the deactivation of
    part of the solid, and compute for solving the equation for a given period
    of time. This class is suited for simulations involving caloric systems
    such as magnetocaloric or electrocaloric systems.

    """

    def __init__(self, amb_temperature, materials=('Cu',), borders=(1, 11),
                 materials_order=(0,), dx=0.01, dt=0.1, file_name=None,
                 boundaries=(0, 0), initial_state=False,
                 materials_path=False, draw=['temperature'], draw_scale=None):
        """Object initialization.

        amb_temperature: ambient temperature of the whole system
        materials: list of strings of all the used materials present in the
            folder materials
        borders: list of the points where there is a change of material
        materials_order: list of the materials list indexes that defines the
            material properties given by borders
        dx: the space step
        dt: the times step
        file_name: file name where the temperature and heat flux are saved
        boundaries: list of two entries that define the boundary condition
            for tempreture. If 0 the boundary condition is insulation
        Q: list of 3 entry lists that gives the fixed heat source coeficient.
            The first term is the initial space index where it is applies. The
            second is the final space index where it is applies. The third is
            the value of the coeficient.
        Q0 is a list of 3 entry lists that gives the temperature dependent heat
            source coefficient. The first term is the initial space index where
            it is applies. The second is the final space index where it is
            applies. The third is the value of the coeficient.
        heat_points: list of the space indexes where we want to extract the
            heat flux. Normally, the first term is the heat flux of the hot end
            and the second term is the heat flux of the cold end
        initial_state: initial state of the materials. True if applied field
            and False is removed field.
        h_left: left heat transfer coefficient
        h_right: right heat transfer coefficient
        materials_path: absolute path of the mateirals database. If false, then
            the materials database is the standard heatrapy database.
        draw: list of strings representing the online plots. If the list is
            empty, then no drawing is performed.
        draw_scale: list of two values, representing the minimum and maximum
            temperature to be drawn. If None, there are no limits.

        """
        # check the validity of inputs
        materials = tuple(materials)
        borders = tuple(borders)
        materials_order = tuple(materials_order)
        boundaries = tuple(boundaries)
        cond01 = isinstance(amb_temperature, float)
        cond01 = cond01 or isinstance(amb_temperature, int)
        cond02 = isinstance(materials, tuple)
        cond03 = isinstance(borders, tuple)
        cond04 = isinstance(materials_order, tuple)
        cond05 = isinstance(dx, int) or isinstance(dx, float)
        cond06 = isinstance(dt, int) or isinstance(dt, float)
        cond07 = isinstance(file_name, str)
        cond07 = cond07 or (file_name is None)
        cond08 = isinstance(boundaries, tuple)
        cond10 = isinstance(initial_state, bool)
        if isinstance(draw, list):
            cond15 = True
        elif draw is None:
            cond15 = True
        else:
            cond15 = False
        if isinstance(draw_scale, list) or isinstance(draw_scale, tuple):
            cond16 = (len(draw_scale) == 2)
        elif draw_scale is None:
            cond16 = True
        else:
            cond16 = False
        condition = cond01 and cond02 and cond03 and cond04 and cond05
        condition = condition and cond06 and cond07 and cond08
        condition = condition and cond10
        condition = condition and cond15 and cond16
        if not condition:
            raise ValueError

        self.object = Object(amb_temperature, materials=materials,
                             borders=borders, materials_order=materials_order,
                             dx=dx, dt=dt, file_name=file_name,
                             boundaries=boundaries,
                             initial_state=initial_state,
                             materials_path=materials_path)

        # initializes the plotting
        self.draw = draw
        self.draw_scale = draw_scale
        for drawing in self.draw:
            if drawing == 'temperature':
                self.figure = plt.figure()
                self.ax = self.figure.add_subplot(111)
                temp = []
                for i in range(len(self.object.temperature)):
                    temp.append(self.object.temperature[i][0])
                if not self.draw_scale:
                    vmax = max(temp)
                    vmin = min(temp)
                    if vmax == vmin:
                        vmin = vmin - 0.1
                        vmax = vmax + 0.1
                    temp = np.array(temp)
                    x_plot = [self.object.dx*j for j in range(len(temp))]
                    self.online, = self.ax.plot(x_plot, temp)
                    self.ax.set_ylim([vmin, vmax])
                else:
                    temp = np.array(temp)
                    x_plot = [self.object.dx*j for j in range(len(temp))]
                    self.online, = self.ax.plot(x_plot, temp)
                    self.ax.set_ylim(self.draw_scale)
                self.ax.set_title('Temperature (K)')
                self.ax.set_xlabel('x axis (m)')
                self.ax.set_ylabel('temperature (K)')
                plt.show(block=False)

    def show_figure(self, figure_type, draw_scale=None):
        """Plotting.

        Initializes a specific plotting. figure_type is a string identifying
        the plotting. draw_scale defines the range of temperatures. If None,
        this range is found automatically for every frame.

        """
        # check the validity of inputs
        if isinstance(draw_scale, list) or isinstance(draw_scale, tuple):
            condition = (len(draw_scale) == 2)
        elif draw_scale is None:
            condition = True
        else:
            condition = False
        condition = condition and isinstance(figure_type, str)
        if not condition:
            raise ValueError

        self.draw_scale = draw_scale
        if figure_type == 'temperature':
            if figure_type not in self.draw:
                self.draw.append(figure_type)
            self.figure = plt.figure()
            self.ax = self.figure.add_subplot(111)
            temp = []
            for i in range(len(self.object.temperature)):
                temp.append(self.object.temperature[i][0])
            if not self.draw_scale:
                vmax = max(temp)
                vmin = min(temp)
                if vmax == vmin:
                    vmin = vmin - 0.1
                    vmax = vmax + 0.1
                temp = np.array(temp)
                x_plot = [self.object.dx*j for j in range(len(temp))]
                self.online, = self.ax.plot(x_plot, temp)
                self.ax.set_ylim([vmin, vmax])
            else:
                temp = np.array(temp)
                x_plot = [self.object.dx*j for j in range(len(temp))]
                self.online, = self.ax.plot(x_plot, temp)
                self.ax.set_ylim(self.draw_scale)
            self.ax.set_title('Temperature (K)')
            self.ax.set_xlabel('x axis (m)')
            self.ax.set_ylabel('temperature (K)')
            plt.show(block=False)

    def activate(self, initial_point, final_point):
        """Activation.

        Activates the thermal object between initial_point to final_point.
        """
        # check the validity of inputs
        condition = isinstance(initial_point, int)
        condition = condition and isinstance(final_point, int)
        if not condition:
            raise ValueError

        self.object.activate(initial_point, final_point)

        if self.draw:
            for drawing in self.draw:
                if drawing == 'temperature':
                    try:
                        temp = []
                        for i in range(len(self.object.temperature)):
                            temp.append(self.object.temperature[i][0])
                        if not self.draw_scale:
                            vmax = max(temp)
                            vmin = min(temp)
                            if vmax == vmin:
                                vmin = vmin - 0.1
                                vmax = vmax + 0.1
                            temp = np.array(temp)
                            self.online.set_ydata(temp)
                            self.ax.set_ylim([vmin, vmax])
                        else:
                            temp = np.array(temp)
                            self.online.set_ydata(temp)
                        self.figure.canvas.draw()
                    except:
                        pass

    def deactivate(self, initial_point, final_point):
        """Deactivation.

        Deactivates the thermal object between initial_point to final_point.
        """
        # check the validity of inputs
        condition = isinstance(initial_point, int)
        condition = condition and isinstance(final_point, int)
        if not condition:
            raise ValueError

        self.object.deactivate(initial_point, final_point)

        if self.draw:
            for drawing in self.draw:
                if drawing == 'temperature':
                    try:
                        temp = []
                        for i in range(len(self.object.temperature)):
                            temp.append(self.object.temperature[i][0])
                        if not self.draw_scale:
                            vmax = max(temp)
                            vmin = min(temp)
                            if vmax == vmin:
                                vmin = vmin - 0.1
                                vmax = vmax + 0.1
                            temp = np.array(temp)
                            self.online.set_ydata(temp)
                            self.ax.set_ylim([vmin, vmax])
                        else:
                            temp = np.array(temp)
                            self.online.set_ydata(temp)
                        self.figure.canvas.draw()
                    except:
                        pass

    def change_power(self, power_type, power, initial_point, final_point):
        """Heat power source change.

        Changes the coeficients for the heat power sources by a value of power
        from initial_point to final_point. power_type is a string that
        represents the type of coefficient, i.e. 'Q' or 'Q0'.

        """
        # check the validity of inputs
        value = isinstance(initial_point, int)
        if value and isinstance(final_point, int):
            cond1 = True
        else:
            cond1 = False
        cond2 = isinstance(power, int) or isinstance(power, float)
        if isinstance(power_type, str):
            if power_type == 'Q' or power_type == 'Q0':
                cond3 = True
            else:
                cond3 = False
        else:
            cond3 = False
        if not (cond1 and cond2 and cond3):
            raise ValueError

        if power_type == 'Q':
            for j in range(initial_point, final_point):
                self.object.Q[j] = power
        if power_type == 'Q0':
            for j in range(initial_point, final_point):
                self.object.Q0[j] = power

    def change_boundaries(self, boundaries):
        """Boundary change.

        Changes boundaries variable.

        """
        # check the validity of inputs
        if isinstance(boundaries, tuple):
            if len(boundaries) == 2:
                condition = True
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.object.boundaries = boundaries

    def compute(self, time_interval, write_interval, solver='explicit_k(x)',
                verbose=True):
        """Compute the thermal process.

        Computes the system for timeInterval, and writes into the file_name
        file every write_interval time steps. Four different solvers can be
        used: 'explicit_general', 'explicit_k(x)', 'implicit_general',
        and 'implicit_k(x)'. If verbose = True, then the progress of the
        computation is shown.

        """
        # check the validity of inputs
        cond1 = isinstance(time_interval, float)
        cond1 = cond1 or isinstance(time_interval, int)
        cond2 = isinstance(write_interval, int)
        if isinstance(solver, str):
            all_solvers = ['implicit_general', 'implicit_k(x)',
                           'explicit_k(x)', 'explicit_general']
            if solver in all_solvers:
                cond3 = True
            else:
                cond3 = False
        else:
            cond3 = False
        cond4 = isinstance(verbose, bool)
        condition = cond1 and cond2 and cond3 and cond4
        if not condition:
            raise ValueError

        # number of time steps for the given timeInterval
        nt = int(time_interval / self.object.dt)

        # number of time steps counting from the last writing process
        nw = 0

        # computes
        for j in range(nt):

            # updates the time_passed
            self.object.time_passed = self.object.time_passed + self.object.dt

            # defines the material properties accoring to the state list
            for i in range(1, self.object.num_points - 1):
                if self.object.state[i] is True:
                    value = self.object.materials_index[i]
                    self.object.rho[i] = self.object.materials[value].rhoa(
                        self.object.temperature[i][0])
                    self.object.Cp[i] = self.object.materials[value].cpa(
                        self.object.temperature[i][0])
                    self.object.k[i] = self.object.materials[value].ka(
                        self.object.temperature[i][0])
                if self.object.state[i] is False:
                    value = self.object.materials_index[i]
                    self.object.rho[i] = self.object.materials[value].rho0(
                        self.object.temperature[i][0])
                    self.object.Cp[i] = self.object.materials[value].cp0(
                        self.object.temperature[i][0])
                    self.object.k[i] = self.object.materials[value].k0(
                        self.object.temperature[i][0])

            # SOLVERS
            # implicit k constant
            if solver == 'implicit_general':
                value = solvers.implicit_general(self.object)
                self.object.temperature, self.object.lheat = value
            # implicit k dependent on x
            if solver == 'implicit_k(x)':
                value = solvers.implicit_k(self.object)
                self.object.temperature, self.object.lheat = value
            # explicit k constant
            if solver == 'explicit_general':
                value = solvers.explicit_general(self.object)
                self.object.temperature, self.object.lheat = value
            # explicit k dependent on x
            if solver == 'explicit_k(x)':
                value = solvers.explicit_k(self.object)
                self.object.temperature, self.object.lheat = value

            nw = nw + 1

            if self.draw:
                for drawing in self.draw:
                    if drawing == 'temperature':
                        try:
                            value = nw + 1 == write_interval
                            if value or j == 0 or j == nt - 1:
                                temp = []
                                for i in range(len(self.object.temperature)):
                                    temp.append(self.object.temperature[i][0])
                                if not self.draw_scale:
                                    vmax = max(temp)
                                    vmin = min(temp)
                                    if vmax == vmin:
                                        vmin = vmin - 0.1
                                        vmax = vmax + 0.1
                                    temp = np.array(temp)
                                    self.online.set_ydata(temp)
                                    self.ax.set_ylim([vmin, vmax])
                                else:
                                    temp = np.array(temp)
                                    self.online.set_ydata(temp)
                                self.figure.canvas.draw()
                        except:
                            pass

            # writes the temperature to file_name file ...
            # if the number of time steps is verified
            if self.object.file_name:
                if nw == write_interval or j == 0 or j == nt - 1:
                    line = '%f,' % self.object.time_passed
                    for i in self.object.temperature:
                        new_line = '%f,' % i[1]
                        line = line + new_line
                    line = line[:-1] + '\n'
                    f = open(self.object.file_name, 'a')
                    f.write(line)
                    f.close()

            if nw == write_interval:
                nw = 0
                if verbose:
                    print('pogress:', int(100*j/nt), '%', end="\r")

        if verbose:
            print('Finished simulation')



"""Contains the classes system_objects and single_object.

Used to compute two-dimensional system models

"""

import copy
from .. import solvers
from . import Object
import numpy as np
import matplotlib.pyplot as plt
import matplotlib


class SingleObject(Object):
    """Single_object class.

    This class solves numerically the two-dimensional heat conduction equation.

    """

    def __init__(self, amb_temperature, material='Cu', dx=0.01, dy=0.01,
                 dt=0.1, size=(10, 10), file_name=None,
                 boundaries=(0, 0, 0, 0), Q=[], Q0=[], initial_state=False,
                 materials_path=False,
                 draw=['temperature', 'materials'], draw_scale=None):
        """Object initialization.

        amb_temperature: ambient temperature of the whole system
        material: background material from materials_path
        dx: space step in the x-axis
        dy: space step in the y-axis
        dt: the times step
        size: tuple of size of the 2D thermal object
        file_name: file name where the temperature and heat flux are saved
        boundaries: list of four entries that define the boundary condition
            for temperature (left, right, bottom, top). If 0 the boundary
            condition is insulation
        Q: list in the form of matrix of fixed heat source coefficient.
        Q0: list in the form of matrix of temperature dependent heat source
            coefficient.
        initial_state: initial state of the materials. True if applied field
            and False is removed field.
        materials_path: absolute path of the materials database.
        draw: list of strings representing the live plots, which can be
            'temperature', 'state', 'materials', 'Q' and 'Q0'.
        draw_scale: list of length 2 representing the minimum and maximum
            temperatures of the temperature plot.

        """
        boundaries = tuple(boundaries)
        cond01 = isinstance(amb_temperature, float)
        cond01 = cond01 or isinstance(amb_temperature, int)
        cond02 = isinstance(material, str)
        cond05 = isinstance(dx, int) or isinstance(dx, float)
        cond06 = isinstance(dy, int) or isinstance(dy, float)
        cond07 = isinstance(dt, int) or isinstance(dt, float)
        cond08 = isinstance(file_name, str)
        cond08 = cond08 or (file_name is None)
        cond09 = isinstance(boundaries, tuple)
        cond10 = isinstance(initial_state, bool)
        cond11 = isinstance(Q, list)
        cond12 = isinstance(Q0, list)
        cond13 = isinstance(draw, list)
        cond14 = isinstance(draw_scale, list) or isinstance(draw_scale, tuple)
        cond14 = cond14 or draw_scale is None
        condition = cond01 and cond02 and cond05
        condition = condition and cond06 and cond07 and cond08 and cond09
        condition = condition and cond10 and cond11 and cond12 and cond13
        condition = condition and cond14
        if not condition:
            raise ValueError

        self.materials_path = materials_path
        self.time_passed = 0.
        self.size = size
        self.dt = dt
        self.dx = dx
        self.dy = dy
        self.object = Object(amb_temperature, material=material, dx=dx, dy=dy,
                             dt=dt, size=size, file_name=file_name,
                             boundaries=boundaries, Q=[], Q0=[],
                             initial_state=initial_state,
                             materials_path=materials_path)

        # Initializes plotting
        self.draw = draw
        self.draw_scale = draw_scale
        if self.draw:
            cmap = matplotlib.cm.RdBu
            name = 'my_cmap_r'
            reverse = []
            k = []
            for key in cmap._segmentdata:
                k.append(key)
                channel = cmap._segmentdata[key]
                data = []
                for t in channel:
                    data.append((1-t[0], t[2], t[1]))
                reverse.append(sorted(data))
            linear = dict(zip(k, reverse))
            my_cmap_r = matplotlib.colors.LinearSegmentedColormap(name, linear)
            cmap_r = my_cmap_r
            for drawing in self.draw:
                if drawing == 'temperature':
                    self.figure = plt.figure()
                    self.ax = self.figure.add_subplot(111)
                    temp = []
                    for i in range(self.object.size[0]):
                        temp.append([])
                        for j in range(self.object.size[1]):
                            temp[-1].append(self.object.temperature[i][j][0])
                    if not self.draw_scale:
                        vmax = max(max(temp, key=max))
                        vmin = min(min(temp, key=min))
                        temp = np.array(temp)
                        extent = [0, size[0]*dx, 0, size[1]*dy]
                        self.im = self.ax.imshow(np.transpose(temp), vmax=vmax,
                                                 vmin=vmin, cmap=cmap_r,
                                                 extent=extent, origin='lower',
                                                 interpolation='hamming')
                    else:
                        temp = np.array(temp)
                        extent = [0, size[0]*dx, 0, size[1]*dy]
                        self.im = self.ax.imshow(np.transpose(temp),
                                                 vmax=self.draw_scale[0],
                                                 vmin=self.draw_scale[1],
                                                 cmap=cmap_r, extent=extent,
                                                 origin='lower',
                                                 interpolation='hamming')
                    cbar_kw = {}
                    cbar = self.ax.figure.colorbar(self.im, ax=self.ax,
                                                   **cbar_kw)
                    self.ax.set_title('Temperature (K)')
                    self.ax.set_xlabel('x axis (m)')
                    self.ax.set_ylabel('y axis (m)')
                    plt.show(block=False)
                if drawing == 'state':
                    self.figure_state = plt.figure()
                    self.ax_state = self.figure_state.add_subplot(111)
                    vmax = 1
                    vmin = 0
                    state = np.array(self.object.state)
                    cmap = plt.get_cmap("gray", 2)
                    extent = [0, size[0]*dx, 0, size[1]*dy]
                    self.im_state = self.ax_state.imshow(np.transpose(state),
                                                         vmax=1.5, vmin=-0.5,
                                                         cmap=cmap,
                                                         extent=extent,
                                                         origin='lower')
                    cbar_kw = {}
                    cbarlabel = ""
                    qrates = np.array(['active', 'inactive'])
                    value = np.linspace(0, 1, 2)
                    norm = matplotlib.colors.BoundaryNorm(value, 1)
                    func = lambda x, pos: qrates[::-1][norm(x)]
                    fmt = matplotlib.ticker.FuncFormatter(func)
                    cbar_kw = dict(ticks=np.arange(-3, 4), format=fmt)
                    cbar_stat = self.ax_state.figure.colorbar(self.im_state,
                                                              ax=self.ax_state,
                                                              **cbar_kw)
                    cbar_stat.ax.set_ylabel(cbarlabel, rotation=-90,
                                            va="bottom")
                    self.ax_state.set_title('State')
                    self.ax_state.set_xlabel('x axis (m)')
                    self.ax_state.set_ylabel('y axis (m)')
                    plt.show(block=False)
                value = len(self.object.materials_name) > 1
                if drawing == 'materials' and value:
                    self.figure_materials = plt.figure()
                    self.ax_materials = self.figure_materials.add_subplot(111)
                    vmax = len(self.object.materials)-1
                    vmin = 0
                    material_id = np.array(self.object.materials_index)
                    cmap = plt.get_cmap("gray", vmax)
                    extent = [0, size[0]*dx, 0, size[1]*dy]
                    origin = 'lower'
                    value = np.transpose(material_id)
                    self.im_materials = self.ax_materials.imshow(value,
                                                                 vmax=vmax,
                                                                 vmin=vmin,
                                                                 cmap=cmap,
                                                                 extent=extent,
                                                                 origin=origin)
                    cbar_kw = {}
                    cbarlabel = ""
                    qrates = np.array(self.object.materials_name)
                    value_1 = np.linspace(0, 1, len(self.object.materials))
                    value_2 = len(self.object.materials)-1
                    norm = matplotlib.colors.BoundaryNorm(value_1, value_2)
                    func = lambda x, pos: qrates[::-1][norm(x)]
                    fmt = matplotlib.ticker.FuncFormatter(func)
                    cbar_kw = dict(ticks=np.arange(-3, 4), format=fmt)
                    value_1 = self.im_materials
                    value_2 = self.ax_materials
                    cbar_m = self.ax_materials.figure.colorbar(value_1,
                                                               ax=value_2,
                                                               **cbar_kw)
                    cbar_m.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
                    self.ax_materials.set_title('Materials')
                    self.ax_materials.set_xlabel('x axis (m)')
                    self.ax_materials.set_ylabel('y axis (m)')
                    plt.show(block=False)
                if drawing == 'Q':
                    self.figure_Q = plt.figure()
                    self.ax_Q = self.figure_Q.add_subplot(111)
                    temp = np.array(self.object.Q)
                    vmax = 1
                    vmin = 0
                    extent = [0, size[0]*dx, 0, size[1]*dy]
                    self.im_Q = self.ax_Q.imshow(np.transpose(temp), vmax=vmax,
                                                 vmin=vmin, cmap='inferno',
                                                 extent=extent, origin='lower',
                                                 interpolation='hamming')
                    cbar_kw = {}
                    cbar = self.ax_Q.figure.colorbar(self.im_Q, ax=self.ax_Q,
                                                     **cbar_kw)
                    self.ax_Q.set_title('Q (W/m²)')
                    self.ax_Q.set_xlabel('x axis (m)')
                    self.ax_Q.set_ylabel('y axis (m)')
                    plt.show(block=False)
                if drawing == 'Q0':
                    self.figure_Q0 = plt.figure()
                    self.ax_Q0 = self.figure_Q0.add_subplot(111)
                    temp = np.array(self.object.Q0)
                    vmax = 1
                    vmin = 0
                    extent = [0, size[0]*dx, 0, size[1]*dy]
                    self.im_Q0 = self.ax_Q0.imshow(np.transpose(temp),
                                                   vmax=vmax, vmin=vmin,
                                                   cmap='inferno',
                                                   extent=extent,
                                                   origin='lower',
                                                   interpolation='hamming')
                    cbar_kw = {}
                    cbar = self.ax_Q0.figure.colorbar(self.im_Q0,
                                                      ax=self.ax_Q0,
                                                      **cbar_kw)
                    self.ax_Q0.set_title('Q0 (W/m²)')
                    self.ax_Q0.set_xlabel('x axis (m)')
                    self.ax_Q0.set_ylabel('y axis (m)')
                    plt.show(block=False)

    def show_figure(self, figure_type):
        """Plotting.

        Initializes a specific plotting. figure_type is a string identifying
        the plotting.

        """
        # check the validity of inputs
        condition = isinstance(figure_type, str)
        if not condition:
            raise ValueError

        if figure_type not in self.draw:
            self.draw.append(figure_type)
        if figure_type == 'temperature':
            self.figure = plt.figure()
            self.ax = self.figure.add_subplot(111)
            temp = []
            for i in range(self.object.size[0]):
                temp.append([])
                for j in range(self.object.size[1]):
                    temp[-1].append(self.object.temperature[i][j][0])
            if not self.draw_scale:
                vmax = max(max(temp, key=max))
                vmin = min(min(temp, key=min))
                temp = np.array(temp)
                extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
                self.im = self.ax.imshow(np.transpose(temp), vmax=vmax,
                                         vmin=vmin, cmap='jet', extent=extent,
                                         origin='lower',
                                         interpolation='hamming')
            else:
                temp = np.array(temp)
                extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
                self.im = self.ax.imshow(np.transpose(temp),
                                         vmax=self.draw_scale[0],
                                         vmin=self.draw_scale[1],
                                         cmap='jet', extent=extent,
                                         origin='lower',
                                         interpolation='hamming')
            cbar_kw = {}
            cbarlabel = "temperature (K)"
            cbar = self.ax.figure.colorbar(self.im, ax=self.ax, **cbar_kw)
            cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
            self.ax.set_title('Temperature')
            self.ax.set_xlabel('x axis (m)')
            self.ax.set_ylabel('y axis (m)')
            plt.show(block=False)
        if figure_type == 'state':
            self.figure_state = plt.figure()
            self.ax_state = self.figure_state.add_subplot(111)
            vmax = 1
            vmin = 0
            state = np.array(self.object.state)
            extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
            self.im_state = self.ax_state.imshow(np.transpose(state), vmax=1.5,
                                                 vmin=-0.5,
                                                 cmap=plt.get_cmap("gray", 2),
                                                 extent=extent, origin='lower')
            cbar_kw = {}
            cbarlabel = ""  # "state"
            qrates = np.array(['active', 'inactive'])
            norm = matplotlib.colors.BoundaryNorm(np.linspace(0, 1, 2), 1)
            func = lambda x, pos: qrates[::-1][norm(x)]
            fmt = matplotlib.ticker.FuncFormatter(func)
            cbar_kw = dict(ticks=np.arange(-3, 4), format=fmt)
            cbar_state = self.ax_state.figure.colorbar(self.im_state,
                                                       ax=self.ax_state,
                                                       **cbar_kw)
            cbar_state.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
            self.ax_state.set_title('State')
            self.ax_state.set_xlabel('x axis (m)')
            self.ax_state.set_ylabel('y axis (m)')
            plt.show(block=False)
        if figure_type == 'materials' and len(self.object.materials_name) > 1:
            self.figure_materials = plt.figure()
            self.ax_materials = self.figure_materials.add_subplot(111)
            vmax = len(self.object.materials)-1
            vmin = 0
            material_id = np.array(self.object.materials_index)
            cmap = plt.get_cmap("PiYG", vmax+1)
            extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
            value = np.transpose(material_id)
            self.im_materials = self.ax_materials.imshow(value,
                                                         vmax=vmax,
                                                         vmin=vmin,
                                                         cmap=cmap,
                                                         extent=extent,
                                                         origin='lower')
            cbar_kw = {}
            cbarlabel = ""
            materials_name_list = copy.deepcopy(self.object.materials_name)
            materials_name_list.reverse()
            qrates = np.array(materials_name_list)
            value = np.linspace(0, len(self.object.materials) - 1,
                                len(self.object.materials))
            norm = matplotlib.colors.BoundaryNorm(value,
                                                  len(self.object.materials)-1)
            func = lambda x, pos: qrates[::-1][norm(x)]
            fmt = matplotlib.ticker.FuncFormatter(func)
            cbar_kw = dict(ticks=np.arange(0, len(self.object.materials)+1),
                           format=fmt)
            cbar_m = self.ax_materials.figure.colorbar(self.im_materials,
                                                       ax=self.ax_materials,
                                                       **cbar_kw)
            cbar_m.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
            self.ax_materials.set_title('Materials')
            self.ax_materials.set_xlabel('x axis (m)')
            self.ax_materials.set_ylabel('y axis (m)')
            plt.show(block=False)
            print(self.object.materials_name)
        if figure_type == 'materials' and len(self.object.materials_name) == 1:
            print(self.object.materials_name)
        if figure_type == 'Q':
            self.figure_Q = plt.figure()
            self.ax_Q = self.figure_Q.add_subplot(111)
            temp = np.array(self.object.Q)
            vmax = 1
            vmin = 0
            extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
            self.im_Q = self.ax_Q.imshow(np.transpose(temp), vmax=vmax,
                                         vmin=vmin, cmap='inferno',
                                         extent=extent, origin='lower',
                                         interpolation='hamming')
            cbar_kw = {}
            cbar = self.ax_Q.figure.colorbar(self.im_Q, ax=self.ax_Q,
                                             **cbar_kw)
            self.ax_Q.set_title('Q (W/m²)')
            self.ax_Q.set_xlabel('x axis (m)')
            self.ax_Q.set_ylabel('y axis (m)')
            plt.show(block=False)
        if figure_type == 'Q0':
            self.figure_Q0 = plt.figure()
            self.ax_Q0 = self.figure_Q0.add_subplot(111)
            temp = np.array(self.object.Q0)
            vmax = 1
            vmin = 0
            extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
            self.im_Q0 = self.ax_Q0.imshow(np.transpose(temp), vmax=vmax,
                                           vmin=vmin, cmap='inferno',
                                           extent=extent,
                                           origin='lower',
                                           interpolation='hamming')
            cbar_kw = {}
            cbar = self.ax_Q0.figure.colorbar(self.im_Q0, ax=self.ax_Q0,
                                              **cbar_kw)
            self.ax_Q0.set_title('Q0 (W/m²)')
            self.ax_Q0.set_xlabel('x axis (m)')
            self.ax_Q0.set_ylabel('y axis (m)')
            plt.show(block=False)

    def activate(self, initial_point, final_point, shape='square'):
        """Activation of the material.

        Activates a given piece of material. If shape is 'square', then the
        initial_point is the tuple (x,y) of the bottom left point and the
        final_point is the tuple (x,y) of the top right point. If the shape is
        'circle', the initial_point is the tuple (x,y) of the center of the
        circle and final_point is its radius.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    condition = len(initial_point) == 2
                    condition = condition and len(final_point) == 2
                else:
                    condition = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    condition = len(initial_point) == 2
                else:
                    condition = False
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.object.activate(initial_point=initial_point,
                             final_point=final_point, shape=shape)
        if self.draw:
            for drawing in self.draw:
                if drawing == 'temperature':
                    temp = []
                    for i in range(self.object.size[0]):
                        temp.append([])
                        for j in range(self.object.size[1]):
                            temp[-1].append(self.object.temperature[i][j][0])
                    self.im.set_array(np.transpose(temp))
                    if not self.draw_scale:
                        vmax = max(max(temp, key=max))
                        vmin = min(min(temp, key=min))
                        self.im.set_clim(vmin=vmin)
                        self.im.set_clim(vmax=vmax)
                    self.figure.canvas.draw()
                if drawing == 'state':
                    self.im_state.set_array(np.transpose(self.object.state))
                    self.figure_state.canvas.draw()

    def deactivate(self, initial_point, final_point, shape='square'):
        """Deactivation of the material.

        Deactivates a given piece of material. If shape is 'square', then the
        initial_point is the tuple (x,y) of the bottom left point and the
        final_point is the tuple (x,y) of the top right point. If the shape is
        'circle', the initial_point is the tuple (x,y) of the center of the
        circle and final_point is its radius.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    condition = len(initial_point) == 2
                    condition = condition and len(final_point) == 2
                else:
                    condition = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    condition = len(initial_point) == 2
                else:
                    condition = False
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.object.deactivate(initial_point=initial_point,
                               final_point=final_point, shape=shape)
        if self.draw:
            for drawing in self.draw:
                if drawing == 'temperature':
                    temp = []
                    for i in range(self.object.size[0]):
                        temp.append([])
                        for j in range(self.object.size[1]):
                            temp[-1].append(self.object.temperature[i][j][0])
                    self.im.set_array(np.transpose(temp))
                    if not self.draw_scale:
                        vmax = max(max(temp, key=max))
                        vmin = min(min(temp, key=min))
                        self.im.set_clim(vmin=vmin)
                        self.im.set_clim(vmax=vmax)
                    self.figure.canvas.draw()
                if drawing == 'state':
                    self.im_state.set_array(np.transpose(self.object.state))
                    self.figure_state.canvas.draw()

    def change_boundaries(self, boundaries):
        """Boundary change.

        Changes boundaries variable.

        """
        # check the validity of inputs
        if isinstance(boundaries, tuple):
            if len(boundaries) == 4:
                condition = True
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.object.boundaries = boundaries

    def change_material(self, shape, material, initial_point, length,
                        state=False):
        """Material change.

        Changes the material of a given piece of the background material. If
        shape is 'square', then the initial_point is the tuple (x,y) of the
        bottom left point and the length is the tuple (x,y) of the length. If
        the shape is 'circle', the initial_point is the tuple (x,y) of the
        center of the circle and length is its radius.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(length, tuple):
                    condition = len(initial_point) == 2
                    condition = condition and len(length) == 2
                else:
                    condition = False
            elif shape == 'circle':
                value = isinstance(length, int)
                value = value or isinstance(length, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    condition = len(initial_point) == 2
                else:
                    condition = False
            else:
                condition = False
        else:
            condition = False
        condition = condition and isinstance(state, bool)
        if not condition:
            raise ValueError

        if shape == 'square':
            self.object.square(material=material,
                               initial_point=initial_point, length=length,
                               state=state, materials_path=self.materials_path)
        if shape == 'circle':
            self.object.circle(material=material,
                               initial_point=initial_point, radius=length,
                               state=state, materials_path=self.materials_path)
        try:
            plt.close(self.figure_materials)
        except:
            pass

        if self.draw:
            if 'materials' in self.draw:
                self.figure_materials = plt.figure()
                self.ax_materials = self.figure_materials.add_subplot(111)
                vmax = len(self.object.materials)-1
                vmin = 0
                cmap = plt.get_cmap("PiYG", vmax+1)
                extent = [0, self.size[0]*self.dx, 0, self.size[1]*self.dy]
                material_id = np.array(self.object.materials_index)
                self.im_materials = self.ax_materials.imshow(np.transpose(material_id),
                                                             vmax=vmax,
                                                             vmin=vmin,
                                                             cmap=cmap,
                                                             extent=extent,
                                                             origin='lower')
                cbar_kw = {}
                cbarlabel = ""
                materials_name_list = copy.deepcopy(self.object.materials_name)
                materials_name_list.reverse()
                qrates = np.array(materials_name_list)
                value_1 = np.linspace(0, len(self.object.materials)-1,
                                      len(self.object.materials))
                value_2 = len(self.object.materials)-1
                norm = matplotlib.colors.BoundaryNorm(value_1, value_2)
                func = lambda x, pos: qrates[::-1][norm(x)]
                fmt = matplotlib.ticker.FuncFormatter(func)
                value = np.arange(0, len(self.object.materials)+1)
                cbar_kw = dict(ticks=value, format=fmt)
                cba_m = self.ax_materials.figure.colorbar(self.im_materials,
                                                          ax=self.ax_materials,
                                                          **cbar_kw)
                cba_m.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
                self.ax_materials.set_title('Materials')
                self.ax_materials.set_xlabel('x axis (m)')
                self.ax_materials.set_ylabel('y axis (m)')
                plt.show(block=False)

    def change_power(self, shape, power_type, initial_point, final_point,
                     power):
        """Power change.

        Changes the power matrix of the thermal object. If shape is 'square',
        then the initial_point is the tuple (x,y) of the bottom left point and
        the final_point is the tuple (x,y) of the top right point. If the shape
        is 'circle', the initial_point is the tuple (x,y) of the center of the
        circle and final_point is its radius. power is the value of the power
        to add, and power type is the type of power to be introduced, which has
        the value 'Q' if it is temperature dependent and 'Q0' if it is
        temperature independent.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    cond1 = len(initial_point) == 2
                    cond1 = cond1 and len(final_point) == 2
                else:
                    cond1 = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    cond1 = len(initial_point) == 2
                else:
                    cond1 = False
            else:
                cond1 = False
        else:
            cond1 = False
        cond2 = isinstance(power, int) or isinstance(power, float)
        if isinstance(power_type, str):
            if power_type == 'Q' or power_type == 'Q0':
                cond3 = True
            else:
                cond3 = False
        else:
            cond3 = False
        if not cond1 and cond2 and cond3:
            raise ValueError

        if shape == 'square':
            self.object.power_add(initial_point, final_point, power,
                                  shape=shape, power_type=power_type)
        if shape == 'circle':
            self.object.power_add(initial_point, final_point, power,
                                  shape=shape, power_type=power_type)
        if self.draw:
            if 'Q' in self.draw and power_type == 'Q':
                self.im_Q.set_array(np.transpose(self.object.Q))
                vmax = max(max(self.object.Q, key=max))
                vmin = min(min(self.object.Q, key=min))
                self.im_Q.set_clim(vmin=vmin)
                self.im_Q.set_clim(vmax=vmax)
                self.figure_Q.canvas.draw()
            if 'Q0' in self.draw and power_type == 'Q0':
                self.im_Q0.set_array(np.transpose(self.object.Q0))
                vmax = max(max(self.object.Q0, key=max))
                vmin = min(min(self.object.Q0, key=min))
                self.im_Q0.set_clim(vmin=vmin)
                self.im_Q0.set_clim(vmax=vmax)
                self.figure_Q0.canvas.draw()

    def compute(self, time_interval, write_interval, solver='explicit_general',
                verbose=True):
        """Compute the thermal process.

        Computes the system for time_interval, and writes into the file_name
        file every write_interval time steps. Two different solvers can be
        used: 'explicit_general' and 'explicit_k(x)'. If verbose = True, then
        the progress of the computation is shown.

        """
        # check the validity of inputs
        cond1 = isinstance(time_interval, float)
        cond1 = cond1 or isinstance(time_interval, int)
        cond2 = isinstance(write_interval, int)
        all_solvers = ['explicit_general', 'explicit_k(x)']
        cond3 = isinstance(solver, str) and solver in all_solvers
        cond4 = isinstance(verbose, bool)
        condition = cond1 and cond2 and cond3 and cond4
        if not condition:
            raise ValueError

        # number of time steps for the given timeInterval
        nt = int(time_interval / self.object.dt)

        # number of time steps counting from the last writing process
        nw = 0

        # computes
        for k in range(nt):

            # updates the time_passed
            self.time_passed = self.time_passed + self.object.dt

            # defines the material properties accoring to the state list
            for i in range(self.object.size[0]):
                for j in range(self.object.size[1]):
                    if self.object.state[i][j] is True:
                        ix = self.object.materials_index[i][j]
                        self.object.rho[i][j] = self.object.materials[ix].rhoa(
                            self.object.temperature[i][j][0])
                        self.object.Cp[i][j] = self.object.materials[ix].cpa(
                            self.object.temperature[i][j][0])
                        self.object.k[i][j] = self.object.materials[ix].ka(
                            self.object.temperature[i][j][0])
                    if self.object.state[i][j] is False:
                        ix = self.object.materials_index[i][j]
                        self.object.rho[i][j] = self.object.materials[ix].rho0(
                            self.object.temperature[i][j][0])
                        self.object.Cp[i][j] = self.object.materials[ix].cp0(
                            self.object.temperature[i][j][0])
                        self.object.k[i][j] = self.object.materials[ix].k0(
                            self.object.temperature[i][j][0])

            # SOLVERS
            # explicit k constant
            if solver == 'explicit_general':
                temp = []
                for i in range(self.object.size[0]):
                    temp.append([])
                    for j in range(self.object.size[1]):
                        temp[-1].append(self.object.temperature[i][j][0])
                value = solvers.explicit_general(self.object)
                self.object.temperature, self.object.lheat = value
                temp = []
                for i in range(self.object.size[0]):
                    temp.append([])
                    for j in range(self.object.size[1]):
                        temp[-1].append(self.object.temperature[i][j][0])
            # explicit k constant
            if solver == 'explicit_k(x)':
                temp = []
                for i in range(self.object.size[0]):
                    temp.append([])
                    for j in range(self.object.size[1]):
                        temp[-1].append(self.object.temperature[i][j][0])
                value = solvers.explicit_k(self.object)
                self.object.temperature, self.object.lheat = value
                temp = []
                for i in range(self.object.size[0]):
                    temp.append([])
                    for j in range(self.object.size[1]):
                        temp[-1].append(self.object.temperature[i][j][0])

            nw = nw + 1

            if self.object.file_name:
                if nw + 1 == write_interval or k == 0 or k == nt - 1:
                    line = '%f' % self.time_passed
                    for i in range(self.object.size[0]):
                        for j in range(self.object.size[1]):
                            new_line = ',%f' % self.object.temperature[i][j][1]
                            line = line + new_line
                    f = open(self.object.file_name, 'a')
                    f.write(line+'\n')
                    f.close()
            if self.draw:
                for drawing in self.draw:
                    if drawing == 'temperature':
                        if nw + 1 == write_interval or k == 0 or k == nt - 1:
                            temp = []
                            for i in range(self.object.size[0]):
                                temp.append([])
                                for j in range(self.object.size[1]):
                                    value = self.object.temperature[i][j][0]
                                    temp[-1].append(value)
                            try:
                                self.im.set_array(np.transpose(temp))
                                if not self.draw_scale:
                                    vmax = max(max(temp, key=max))
                                    vmin = min(min(temp, key=min))
                                    self.im.set_clim(vmin=vmin)
                                    self.im.set_clim(vmax=vmax)
                                self.figure.canvas.draw()
                            except:
                                pass

            if nw == write_interval:
                nw = 0
                if verbose:
                    print('pogress:', int(100*k/nt), '%', end="\r")
        if verbose:
            print('Finished simulation')



"""Solvers.

This submodule contains the solvers for the computation of thermal processes.

"""

from .implicit_k import implicit_k
from .implicit_general import implicit_general
from .explicit_k import explicit_k
from .explicit_general import explicit_general

__all__ = [implicit_k, implicit_general, explicit_k, explicit_general]



"""2D Objects.

This submodule includes the object class for creating two-dimensional thermal
objects, object system class for creating systems of two-dimensional thermal
objects, and a complex single object for using in more complex thermal
computing.

"""

from .object import Object
from .system import SystemObjects
from .single import SingleObject

__all__ = [Object, SystemObjects, SingleObject]



"""Contains the class calmatpro.

Used to access the properties of materials from the database

"""

import numpy as np


class CalMatPro:
    """CalMatPro class.

    Loads and gives physical properties of materials from the database

    """

    def __init__(self, name_tadi, name_tadd, name_cpa, name_cp0, name_k0,
                 name_ka, name_rho0, name_rhoa, name_lheat0, name_lheata):
        """Object initialization.

        loads the physical properties of a materials
        nameTadi is the adiabatic temperature increase file
        nameTadd is the adiabatic temperature decrease file
        nameCp0 is the specific heat file
        nameCpA is the active specific heat file
        namek0 is the thermal conductivity file
        namekA is the active thermal conductivity file
        namerho0 is the density file
        namerhoA is the active density file

        """
        # adiabatic temperature increase
        input = open(name_tadi, 'r')
        s = input.readlines()
        self.xTadi = []
        self.yTadi = []
        for line in s:
            pair_tad = line.split()
            self.xTadi.append(float(pair_tad[0]))
            self.yTadi.append(float(pair_tad[1]))
        input.close()

        # adiabatic temperature decrease
        input = open(name_tadd, 'r')
        s = input.readlines()
        self.xTadd = []
        self.yTadd = []
        for line in s:
            pair_tad = line.split()
            self.xTadd.append(float(pair_tad[0]))
            self.yTadd.append(float(pair_tad[1]))
        input.close()

        # specific heat
        input = open(name_cpa, 'r')
        s = input.readlines()
        self.xCpA = []
        self.yCpA = []
        for line in s:
            pair_tad = line.split()
            self.xCpA.append(float(pair_tad[0]))
            self.yCpA.append(float(pair_tad[1]))
        input.close()

        # active specific heat
        input = open(name_cp0, 'r')
        s = input.readlines()
        self.xCp0 = []
        self.yCp0 = []
        for line in s:
            pair_tad = line.split()
            self.xCp0.append(float(pair_tad[0]))
            self.yCp0.append(float(pair_tad[1]))
        input.close()

        # thermal conductivity
        input = open(name_k0, 'r')
        s = input.readlines()
        self.xk0 = []
        self.yk0 = []
        for line in s:
            pair_tad = line.split()
            self.xk0.append(float(pair_tad[0]))
            self.yk0.append(float(pair_tad[1]))
        input.close()

        # active thermal conductivity
        input = open(name_ka, 'r')
        s = input.readlines()
        self.xkA = []
        self.ykA = []
        for line in s:
            pair_tad = line.split()
            self.xkA.append(float(pair_tad[0]))
            self.ykA.append(float(pair_tad[1]))
        input.close()

        # density
        input = open(name_rho0, 'r')
        s = input.readlines()
        self.xrho0 = []
        self.yrho0 = []
        for line in s:
            pair_tad = line.split()
            self.xrho0.append(float(pair_tad[0]))
            self.yrho0.append(float(pair_tad[1]))
        input.close()

        # active density
        input = open(name_rhoa, 'r')
        s = input.readlines()
        self.xrhoA = []
        self.yrhoA = []
        for line in s:
            pair_tad = line.split()
            self.xrhoA.append(float(pair_tad[0]))
            self.yrhoA.append(float(pair_tad[1]))
        input.close()

        # latent heat
        input = open(name_lheat0, 'r')
        s = input.readlines()
        self.latent_heat0 = []
        for line in s:
            latent_values = line.split()
            self.latent_heat0.append(
                (float(latent_values[0]), float(latent_values[1]))
            )
        input.close()

        # active latent heat
        input = open(name_lheata, 'r')
        s = input.readlines()
        self.latent_heatA = []
        for line in s:
            latent_values = line.split()
            self.latent_heatA.append(
                (float(latent_values[0]),float(latent_values[1]))
            )
        input.close()

    def tadi(self, temperature):
        """Adiabatic temperature increase.

        Gives the adiabatic temperature increase for a given temperature

        """
        return np.interp(temperature, self.xTadi, self.yTadi)

    def tadd(self, temperature):
        """Adiabatic temperature decrease.

        Gives the adiabatic temperature decrease for a given temperature

        """
        return np.interp(temperature, self.xTadd, self.yTadd)

    def cpa(self, temperature):
        """Active specific heat.

        Gives the active specific heat for a given temperature

        """
        return np.interp(temperature, self.xCpA, self.yCpA)

    def cp0(self, temperature):
        """Inactive specific heat.

        Gives the specific heat for a given temperature

        """
        return np.interp(temperature, self.xCp0, self.yCp0)

    def k0(self, temperature):
        """Inactive thermal conductivity.

        Gives the thermal conductivity for a given temperature

        """
        return np.interp(temperature, self.xk0, self.yk0)

    def ka(self, temperature):
        """Active thermal conductivity.

        Gives the active thermal conductivity for a given temperature

        """
        return np.interp(temperature, self.xkA, self.ykA)

    def rho0(self, temperature):
        """Inactive density.

        Gives the density for a given temperature

        """
        return np.interp(temperature, self.xrho0, self.yrho0)

    def rhoa(self, temperature):
        """Active density.

        Gives the active density for a given temperature

        """
        return np.interp(temperature, self.xrhoA, self.yrhoA)

    def lheat0(self):
        """Inactive latent heat.

        Gives the latent heat list

        """
        return self.latent_heat0

    def lheata(self):
        """Active latent heat.

        Gives the active latent heat list

        """
        return self.latent_heatA



"""Contains the class object.

Used to create two-dimensional thermal objects, and apply some methods.

"""

from ... import mats
import os
import copy
import numpy as np


class Object:
    """Object class.

    This class creates thermal objects at 2D to be used in more complex
    systems.

    """

    def __init__(self, amb_temperature, material='Cu', dx=0.01, dy=0.01,
                 dt=0.1, size=(10, 10), file_name=None,
                 boundaries=(0, 0, 0, 0), Q=[], Q0=[], initial_state=False,
                 heat_save=False, materials_path=False):
        """Thermal object initialization.

        amb_temperature: ambient temperature of the whole system
        material: background material present at materials_path
        dx: space step along the x-axis
        dy: space step along the y-axis
        dt: the times step
        file_name: file name where the temperature is saved
        boundaries: list of four entries that define the boundary condition
            for temperature (left, right, bottom, top). If 0 the boundary
            condition is insulation
        Q: list in the form of matrix of fixed heat source coefficient.
        Q0: list in the form of matrix of temperature dependent heat source
            coefficient.
        initial_state: initial state of the materials. True if applied field
            and False is removed field.
        heat_save: True if saving the heat at the two borders.
        materials_path: absolute path of the materials database.

        """
        # check the validity of inputs
        boundaries = tuple(boundaries)
        Q = list(Q)
        Q0 = list(Q0)
        cond01 = isinstance(amb_temperature, float)
        cond01 = cond01 or isinstance(amb_temperature, int)
        cond05 = isinstance(dx, int) or isinstance(dx, float)
        cond06 = isinstance(dt, int) or isinstance(dt, float)
        cond07 = isinstance(file_name, str)
        cond07 = cond07 or (file_name is None)
        cond08 = isinstance(boundaries, tuple)
        cond09 = isinstance(Q, list)
        cond10 = isinstance(Q0, list)
        cond11 = isinstance(initial_state, bool)
        cond12 = isinstance(heat_save, bool)
        condition = cond01 and cond05
        condition = condition and cond06 and cond07 and cond08 and cond09
        condition = condition and cond10 and cond11 and cond12
        if not condition:
            raise ValueError

        self.materials = [material]
        self.materials_name = [material]
        self.boundaries = boundaries
        self.amb_temperature = amb_temperature

        if materials_path is False:
            tadi = os.path.dirname(os.path.realpath(__file__)) + \
                '/../../database/' + self.materials[0] + '/' + 'tadi.txt'
            tadd = os.path.dirname(os.path.realpath(__file__)) + \
                '/../../database/' + self.materials[0] + '/' + 'tadd.txt'
            cpa = os.path.dirname(os.path.realpath(__file__)) + \
                '/../../database/' + self.materials[0] + '/' + 'cpa.txt'
            cp0 = os.path.dirname(os.path.realpath(__file__)) + \
                '/../../database/' + self.materials[0] + '/' + 'cp0.txt'
            k0 = os.path.dirname(os.path.realpath(__file__)) + \
                '/../../database/' + self.materials[0] + '/' + 'k0.txt'
            ka = os.path.dirname(os.path.realpath(__file__)) + \
                '/../../database/' + self.materials[0] + '/' + 'ka.txt'
            rho0 = os.path.dirname(os.path.realpath(__file__)) + \
                '/../../database/' + self.materials[0] + '/' + 'rho0.txt'
            rhoa = os.path.dirname(os.path.realpath(__file__)) + \
                '/../../database/' + self.materials[0] + '/' + 'rhoa.txt'
            lheat0 = os.path.dirname(os.path.realpath(__file__)) + \
                '/../../database/' + self.materials[0] + '/' + 'lheat0.txt'
            lheata = os.path.dirname(os.path.realpath(__file__)) + \
                '/../../database/' + self.materials[0] + '/' + 'lheata.txt'
            self.materials[0] = mats.CalMatPro(
                tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata)
        else:
            tadi = materials_path + self.materials[0] + '/' + 'tadi.txt'
            tadd = materials_path + self.materials[0] + '/' + 'tadd.txt'
            cpa = materials_path + self.materials[0] + '/' + 'cpa.txt'
            cp0 = materials_path + self.materials[0] + '/' + 'cp0.txt'
            k0 = materials_path + self.materials[0] + '/' + 'k0.txt'
            ka = materials_path + self.materials[0] + '/' + 'ka.txt'
            rho0 = materials_path + self.materials[0] + '/' + 'rho0.txt'
            rhoa = materials_path + self.materials[0] + '/' + 'rhoa.txt'
            lheat0 = materials_path + self.materials[0] + '/' + 'lheat0.txt'
            lheata = materials_path + self.materials[0] + '/' + 'lheata.txt'
            self.materials[0] = mats.CalMatPro(
                tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata)

        self.materials_index = [None]
        self.size = size
        self.file_name = file_name
        self.dx = dx
        self.dy = dy
        self.dt = dt
        self.temperature = []
        self.latent_heat = []
        self.lheat = []
        self.Cp = []
        self.rho = []
        self.Q = []
        self.Q0 = []
        self.k = []
        self.materials_index = []
        self.state = []
        for i in range(self.size[0]):
            self.state.append([])
            self.materials_index.append([])
            self.temperature.append([])
            self.latent_heat.append([])
            self.lheat.append([])
            self.Cp.append([])
            self.rho.append([])
            self.Q.append([0. for i in range(self.size[1])])
            self.Q0.append([0. for i in range(self.size[1])])
            self.k.append([])
            for j in range(self.size[1]):
                self.materials_index[-1].append(0)
                self.temperature[-1].append([amb_temperature, amb_temperature])
                self.state[-1].append(initial_state)
                if initial_state:
                    value = self.materials[self.materials_index[i][j]]
                    self.Cp[-1].append(value.cpa(self.amb_temperature))
                    value = self.materials[self.materials_index[i][j]]
                    self.rho[-1].append(value.rhoa(self.amb_temperature))
                    value = self.materials[self.materials_index[i][j]]
                    self.k[-1].append(value.ka(self.amb_temperature))
                    self.latent_heat[-1].append(
                        self.materials[self.materials_index[i][j]].lheata()
                    )
                    self.lheat[-1].append([])
                    value = self.materials[self.materials_index[i][j]]
                    for lh in value.lheata():
                        if self.temperature[i][j][1] < lh[0] and lh[1] > 0.:
                            self.lheat[-1].append([lh[0], 0.])
                        if self.temperature[i][j][1] > lh[0] and lh[1] > 0.:
                            self.lheat[-1].append([lh[0], lh[1]])
                        if self.temperature[i][j][1] < lh[0] and lh[1] < 0.:
                            self.lheat[-1].append([lh[0], -lh[1]])
                        if self.temperature[i][j][1] > lh[0] and lh[1] < 0.:
                            self.lheat[-1].append([lh[0], 0.])
                else:
                    value = self.materials[self.materials_index[i][j]]
                    self.Cp[-1].append(value.cp0(self.amb_temperature))
                    value = self.materials[self.materials_index[i][j]]
                    self.rho[-1].append(value.rho0(self.amb_temperature))
                    value = self.materials[self.materials_index[i][j]]
                    self.k[-1].append(value.k0(self.amb_temperature))
                    self.latent_heat[-1].append(
                        self.materials[self.materials_index[i][j]].lheat0()
                    )
                    self.lheat[-1].append([])
                    value = self.materials[self.materials_index[i][j]]
                    for lh in value.lheat0():
                        if self.temperature[i][j][1] < lh[0] and lh[1] > 0.:
                            self.lheat[-1].append([lh[0], 0.])
                        if self.temperature[i][j][1] > lh[0] and lh[1] > 0.:
                            self.lheat[-1].append([lh[0], lh[1]])
                        if self.temperature[i][j][1] < lh[0] and lh[1] < 0.:
                            self.lheat[-1].append([lh[0], -lh[1]])
                        if self.temperature[i][j][1] > lh[0] and lh[1] < 0.:
                            self.lheat[-1].append([lh[0], 0.])

        if Q != []:
            self.Q = Q
        if Q0 != []:
            self.Q0 = Q0

        self.time_passed = 0.

        self.Q_ref = copy.copy(self.Q)
        self.Q0_ref = copy.copy(self.Q0)

        if file_name:
            line = 'time(s)'
            for i in range(size[0]):
                for j in range(size[1]):
                    line = line + ',T[' + str(i) + '][' + str(j) + '] (K)'
            if heat_save:
                line = line + ',Q (J/m)'
            line = line + '\n'
            f = open(self.file_name, 'a')
            f.write(line)
            f.close()

    def activate(self, initial_point, final_point, shape='square'):
        """Activation of the material.

        Activates a given piece of material. If shape is 'square', then the
        initial_point is the tuple (x,y) of the bottom left point and the
        final_point is the tuple (x,y) of the top right point. If the shape is
        'circle', the initial_point is the tuple (x,y) of the center of the
        circle and final_point is its radius.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    condition = len(initial_point) == 2
                    condition = condition and len(final_point) == 2
                else:
                    condition = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    condition = len(initial_point) == 2
                else:
                    condition = False
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        if shape == 'square':
            initial_point_x = int(initial_point[0])
            initial_point_y = int(initial_point[1])
            final_point_x = int(final_point[0])
            final_point_y = int(final_point[1])
            for i in range(initial_point_x, final_point_x):
                for j in range(initial_point_y, final_point_y):
                    if self.state[i][j] is False:
                        value = self.temperature[i][j][0]
                        self.temperature[i][j][0] = value + \
                            self.materials[self.materials_index[i][j]].tadi(
                                self.temperature[i][j][0])
                        value = self.materials_index[i][j]
                        self.rho[i][j] = self.materials[value].rhoa(
                            self.temperature[i][j][0])
                        self.Cp[i][j] = self.materials[value].cpa(
                            self.temperature[i][j][0])
                        self.k[i][j] = self.materials[value].ka(
                            self.temperature[i][j][0])
                        self.lheat[i][j] = []
                        valh = self.materials[value].lheata()
                        self.latent_heat[i][j] = valh
                        for lh in self.latent_heat[i][j]:
                            cond = self.temperature[i][j][0] < lh[0]
                            if cond and lh[1] > 0.:
                                self.lheat[i][j].append([lh[0], 0.])
                            cond = self.temperature[i][j][0] > lh[0]
                            if cond and lh[1] > 0.:
                                self.lheat[i][j].append([lh[0], lh[1]])
                            cond = self.temperature[i][j][0] < lh[0]
                            if cond and lh[1] < 0.:
                                self.lheat[i][j].append([lh[0], -lh[1]])
                            cond = self.temperature[i][j][0] > lh[0]
                            if cond and lh[1] < 0.:
                                self.lheat[i][j].append([lh[0], 0.])
                        self.state[i][j] = True
                    else:
                        message = 'point ({:d},{:d})'.format(i, j)
                        message = message + ' already activated'
                        print(message)

        if shape == 'circle':
            initial_point_x = int(initial_point[0])
            initial_point_y = int(initial_point[1])
            radius = int(final_point)
            for i in range(self.size[0]):
                for j in range(self.size[1]):
                    value = (i-initial_point_x)**2+(j-initial_point_y)**2
                    length = np.sqrt(value)
                    if length <= radius:
                        if self.state[i][j] is False:
                            value = self.temperature[i][j][0]
                            index_value = self.materials_index[i][j]
                            self.temperature[i][j][0] = value + \
                                self.materials[index_value].tadi(
                                    self.temperature[i][j][0])
                            self.rho[i][j] = self.materials[index_value].rhoa(
                                self.temperature[i][j][0])
                            self.Cp[i][j] = self.materials[index_value].cpa(
                                self.temperature[i][j][0])
                            self.k[i][j] = self.materials[index_value].ka(
                                self.temperature[i][j][0])
                            self.lheat[i][j] = []
                            valh = self.materials[index_value].lheata()
                            self.latent_heat[i][j] = valh
                            for lh in self.latent_heat[i][j]:
                                value = self.temperature[i][j][0] < lh[0]
                                if value and lh[1] > 0.:
                                    self.lheat[i][j].append([lh[0], 0.])
                                value = self.temperature[i][j][0] > lh[0]
                                if value and lh[1] > 0.:
                                    self.lheat[i][j].append([lh[0], lh[1]])
                                value = self.temperature[i][j][0] < lh[0]
                                if value and lh[1] < 0.:
                                    self.lheat[i][j].append([lh[0], -lh[1]])
                                value = self.temperature[i][j][0] > lh[0]
                                if value and lh[1] < 0.:
                                    self.lheat[i][j].append([lh[0], 0.])
                            self.state[i][j] = True
                        else:
                            message = 'point ({:d},{:d})'.format(i, j)
                            message = message + ' already activated'
                            print(message)

    def deactivate(self, initial_point, final_point, shape='square'):
        """Deactivation of the material.

        Deactivates a given piece of material. If shape is 'square', then the
        initial_point is the tuple (x,y) of the bottom left point and the
        final_point is the tuple (x,y) of the top right point. If the shape is
        'circle', the initial_point is the tuple (x,y) of the center of the
        circle and final_point is its radius.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    condition = len(initial_point) == 2
                    condition = condition and len(final_point) == 2
                else:
                    condition = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    condition = len(initial_point) == 2
                else:
                    condition = False
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        if shape == 'square':
            initial_point_x = int(initial_point[0])
            initial_point_y = int(initial_point[1])
            final_point_x = int(final_point[0])
            final_point_y = int(final_point[1])
            for i in range(initial_point_x, final_point_x):
                for j in range(initial_point_y, final_point_y):
                    if self.state[i][j] is True:
                        value = self.temperature[i][j][0]
                        self.temperature[i][j][0] = value - \
                            self.materials[self.materials_index[i][j]].tadd(
                                self.temperature[i][j][0])
                        value = self.materials_index[i][j]
                        self.rho[i][j] = self.materials[value].rho0(
                            self.temperature[i][j][0])
                        self.Cp[i][j] = self.materials[value].cp0(
                            self.temperature[i][j][0])
                        self.k[i][j] = self.materials[value].k0(
                            self.temperature[i][j][0])
                        self.lheat[i][j] = []
                        valh = self.materials[value].lheat0()
                        self.latent_heat[i][j] = valh
                        for lh in self.latent_heat[i][j]:
                            cond = self.temperature[i][j][0] < lh[0]
                            if cond and lh[1] > 0.:
                                self.lheat[i][j].append([lh[0], 0.])
                            cond = self.temperature[i][j][0] > lh[0]
                            if cond and lh[1] > 0.:
                                self.lheat[i][j].append([lh[0], lh[1]])
                            cond = self.temperature[i][j][0] < lh[0]
                            if cond and lh[1] < 0.:
                                self.lheat[i][j].append([lh[0], -lh[1]])
                            cond = self.temperature[i][j][0] > lh[0]
                            if cond and lh[1] < 0.:
                                self.lheat[i][j].append([lh[0], 0.])
                        self.state[i][j] = False
                    else:
                        message = 'point ({:d},{:d})'.format(i, j)
                        message = message + ' already deactivated'
                        print(message)

        if shape == 'circle':
            initial_point_x = int(initial_point[0])
            initial_point_y = int(initial_point[1])
            radius = int(final_point)
            for i in range(self.size[0]):
                for j in range(self.size[1]):
                    value = (i-initial_point_x)**2+(j-initial_point_y)**2
                    length = np.sqrt(value)
                    if length <= radius:
                        if self.state[i][j] is False:
                            value = self.temperature[i][j][0]
                            index_value = self.materials_index[i][j]
                            self.temperature[i][j][0] = value - \
                                self.materials[index_value].tadd(
                                    self.temperature[i][j][0])
                            self.rho[i][j] = self.materials[index_value].rho0(
                                self.temperature[i][j][0])
                            self.Cp[i][j] = self.materials[index_value].cp0(
                                self.temperature[i][j][0])
                            self.k[i][j] = self.materials[index_value].k0(
                                self.temperature[i][j][0])
                            self.lheat[i][j] = []
                            valh = self.materials[index_value].lheat0()
                            self.latent_heat[i][j] = valh
                            for lh in self.latent_heat[i][j]:
                                value = self.temperature[i][j][0] < lh[0]
                                if value and lh[1] > 0.:
                                    self.lheat[i][j].append([lh[0], 0.])
                                value = self.temperature[i][j][0] > lh[0]
                                if value and lh[1] > 0.:
                                    self.lheat[i][j].append([lh[0], lh[1]])
                                value = self.temperature[i][j][0] < lh[0]
                                if value and lh[1] < 0.:
                                    self.lheat[i][j].append([lh[0], -lh[1]])
                                value = self.temperature[i][j][0] > lh[0]
                                if value and lh[1] < 0.:
                                    self.lheat[i][j].append([lh[0], 0.])
                            self.state[i][j] = False
                        else:
                            message = 'point ({:d},{:d})'.format(i, j)
                            message = message + ' already deactivated'
                            print(message)

    def square(self, material='Gd', initial_point=(3, 3), length=(3, 3),
               state=False, materials_path=False):
        """Material adding with rectangle shape.

        Adds a new material with a rectangel shape, where initial_point is the
        bottom left (x,y) tuple, length is the length along the two axis, state
        is the initial state of the material and materials_path is the absolute
        path of the materials database.

        """
        # check the validity of inputs
        value = isinstance(initial_point, tuple)
        if value and isinstance(length, tuple):
            cond1 = len(initial_point) == 2
            cond1 = cond1 and len(length) == 2
        else:
            cond1 = False
        cond2 = isinstance(material, str)
        cond3 = isinstance(state, bool)
        cond4 = isinstance(materials_path, str) or materials_path is False
        if not cond1 and cond2 and cond3 and cond4:
            raise ValueError

        initial_point_x = int(initial_point[0])
        initial_point_y = int(initial_point[1])
        final_point_x = initial_point_x + int(length[0])
        final_point_y = initial_point_y + int(length[1])
        if material in self.materials_name:
            index = self.materials_name.index(material)
        else:
            index = len(self.materials)
            self.materials_name.append(material)
            if materials_path is False:
                value = self.materials_name[index]
                tadi = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'tadi.txt'
                tadd = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'tadd.txt'
                cpa = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'cpa.txt'
                cp0 = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'cp0.txt'
                k0 = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'k0.txt'
                ka = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'ka.txt'
                rho0 = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'rho0.txt'
                rhoa = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'rhoa.txt'
                lheat0 = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'lheat0.txt'
                lheata = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'lheata.txt'
                self.materials.append(mats.CalMatPro(
                    tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata))
            else:
                value = self.materials_name[index]
                tadi = materials_path + value + '/' + 'tadi.txt'
                tadd = materials_path + value + '/' + 'tadd.txt'
                cpa = materials_path + value + '/' + 'cpa.txt'
                cp0 = materials_path + value + '/' + 'cp0.txt'
                k0 = materials_path + value + '/' + 'k0.txt'
                ka = materials_path + value + '/' + 'ka.txt'
                rho0 = materials_path + value + '/' + 'rho0.txt'
                rhoa = materials_path + value + '/' + 'rhoa.txt'
                lheat0 = materials_path + value + '/' + 'lheat0.txt'
                lheata = materials_path + value + '/' + 'lheata.txt'
                self.materials.append(mats.CalMatPro(
                    tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata))

        for i in range(initial_point_x, final_point_x):
            for j in range(initial_point_y, final_point_y):
                if state is False:
                    self.state[i][j] = False
                    self.materials_index[i][j] = index
                    self.rho[i][j] = self.materials[index].rho0(
                        self.temperature[i][j][0])
                    self.Cp[i][j] = self.materials[index].cp0(
                        self.temperature[i][j][0])
                    self.k[i][j] = self.materials[index].k0(
                        self.temperature[i][j][0])
                    self.lheat[i][j] = []
                    valh = self.materials[index].lheat0()
                    self.latent_heat[i][j] = valh
                    for lh in self.latent_heat[i][j]:
                        if self.temperature[i][j][0] < lh[0] and lh[1] > 0:
                            self.lheat[i][j].append([lh[0], 0.])
                        if self.temperature[i][j][0] > lh[0] and lh[1] > 0:
                            self.lheat[i][j].append([lh[0], lh[1]])
                        if self.temperature[i][j][0] < lh[0] and lh[1] < 0:
                            self.lheat[i][j].append([lh[0], -lh[1]])
                        if self.temperature[i][j][0] > lh[0] and lh[1] < 0:
                            self.lheat[i][j].append([lh[0], 0.])

                else:
                    self.state[i][j] = True
                    self.materials_index[i][j] = index
                    self.rho[i][j] = self.materials[index].rhoa(
                        self.temperature[i][j][0])
                    self.Cp[i][j] = self.materials[index].cpa(
                        self.temperature[i][j][0])
                    self.k[i][j] = self.materials[index].ka(
                        self.temperature[i][j][0])
                    self.lheat[i][j] = []
                    valh = self.materials[index].lheata()
                    self.latent_heat[i][j] = valh
                    for lh in self.latent_heat[i][j]:
                        if self.temperature[i][j][0] < lh[0] and lh[1] > 0:
                            self.lheat[i][j].append([lh[0], 0.])
                        if self.temperature[i][j][0] > lh[0] and lh[1] > 0:
                            self.lheat[i][j].append([lh[0], lh[1]])
                        if self.temperature[i][j][0] < lh[0] and lh[1] < 0:
                            self.lheat[i][j].append([lh[0], -lh[1]])
                        if self.temperature[i][j][0] > lh[0] and lh[1] < 0:
                            self.lheat[i][j].append([lh[0], 0.])

    def circle(self, material='Gd', initial_point=(3, 3), radius=3,
               state=False, materials_path=False):
        """Material adding with circle shape.

        Adds a new material with a circle shape, where initial_point is the
        (x,y) tuple of the center of the circle, radius is the radius of the
        circle, state is the initial state of the material and materials_path
        is the absolute path of the materials database.

        """
        # check the validity of inputs
        if isinstance(initial_point, tuple):
            cond1 = len(initial_point) == 2
        else:
            cond1 = False
        cond2 = isinstance(radius, int) or isinstance(radius, float)
        cond3 = isinstance(material, str)
        cond4 = isinstance(state, bool)
        cond5 = isinstance(materials_path, str) or materials_path is False
        if not cond1 and cond2 and cond3 and cond4 and cond5:
            raise ValueError

        initial_point_x = int(initial_point[0])
        initial_point_y = int(initial_point[1])
        if material in self.materials_name:
            index = self.materials_name.index(material)
        else:
            index = len(self.materials)
            self.materials_name.append(material)
            if materials_path is False:
                value = self.materials_name[index]
                tadi = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'tadi.txt'
                tadd = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'tadd.txt'
                cpa = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'cpa.txt'
                cp0 = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'cp0.txt'
                k0 = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'k0.txt'
                ka = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'ka.txt'
                rho0 = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'rho0.txt'
                rhoa = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'rhoa.txt'
                lheat0 = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'lheat0.txt'
                lheata = os.path.dirname(os.path.realpath(__file__)) + \
                    '/../../database/' + value + '/' + 'lheata.txt'
                self.materials.append(mats.CalMatPro(
                    tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata))
            else:
                value = self.materials_name[index]
                tadi = materials_path + value + '/' + 'tadi.txt'
                tadd = materials_path + value + '/' + 'tadd.txt'
                cpa = materials_path + value + '/' + 'cpa.txt'
                cp0 = materials_path + value + '/' + 'cp0.txt'
                k0 = materials_path + value + '/' + 'k0.txt'
                ka = materials_path + value + '/' + 'ka.txt'
                rho0 = materials_path + value + '/' + 'rho0.txt'
                rhoa = materials_path + value + '/' + 'rhoa.txt'
                lheat0 = materials_path + value + '/' + 'lheat0.txt'
                lheata = materials_path + value + '/' + 'lheata.txt'
                self.materials.append(mats.CalMatPro(
                    tadi, tadd, cpa, cp0, k0, ka, rho0, rhoa, lheat0, lheata))

        for i in range(self.size[0]):
            for j in range(self.size[1]):
                length = np.sqrt((i-initial_point_x)**2+(j-initial_point_y)**2)
                if length <= radius:
                    if state is False:
                        self.state[i][j] = False
                        self.materials_index[i][j] = index
                        self.rho[i][j] = self.materials[index].rho0(
                            self.temperature[i][j][0])
                        self.Cp[i][j] = self.materials[index].cp0(
                            self.temperature[i][j][0])
                        self.k[i][j] = self.materials[index].k0(
                            self.temperature[i][j][0])
                        self.lheat[i][j] = []
                        valh = self.materials[index].lheat0()
                        self.latent_heat[i][j] = valh
                        for lh in self.latent_heat[i][j]:
                            if self.temperature[i][j][0] < lh[0] and lh[1] > 0:
                                self.lheat[i][j].append([lh[0], 0.])
                            if self.temperature[i][j][0] > lh[0] and lh[1] > 0:
                                self.lheat[i][j].append([lh[0], lh[1]])
                            if self.temperature[i][j][0] < lh[0] and lh[1] < 0:
                                self.lheat[i][j].append([lh[0], -lh[1]])
                            if self.temperature[i][j][0] > lh[0] and lh[1] < 0:
                                self.lheat[i][j].append([lh[0], 0.])

                    else:
                        self.state[i][j] = True
                        self.materials_index[i][j] = index
                        self.rho[i][j] = self.materials[index].rhoa(
                            self.temperature[i][j][0])
                        self.Cp[i][j] = self.materials[index].cpa(
                            self.temperature[i][j][0])
                        self.k[i][j] = self.materials[index].ka(
                            self.temperature[i][j][0])
                        self.lheat[i][j] = []
                        valh = self.materials[index].lheata()
                        self.latent_heat[i][j] = valh
                        for lh in self.latent_heat[i][j]:
                            if self.temperature[i][j][0] < lh[0] and lh[1] > 0:
                                self.lheat[i][j].append([lh[0], 0.])
                            if self.temperature[i][j][0] > lh[0] and lh[1] > 0:
                                self.lheat[i][j].append([lh[0], lh[1]])
                            if self.temperature[i][j][0] < lh[0] and lh[1] < 0:
                                self.lheat[i][j].append([lh[0], -lh[1]])
                            if self.temperature[i][j][0] > lh[0] and lh[1] < 0:
                                self.lheat[i][j].append([lh[0], 0.])

    def power_add(self, initial_point, final_point, power, shape='square',
                  power_type='Q'):
        """Power adding.

        Adds a power matrix to the thermal object. If shape is 'square', then
        the initial_point is the tuple (x,y) of the bottom left point and the
        final_point is the tuple (x,y) of the top right point. If the shape is
        'circle', the initial_point is the tuple (x,y) of the center of the
        circle and final_point is its radius. power is the value of the power
        to add, and power type is the type of power to be introduced, which has
        the value 'Q' if it is temperature dependent and 'Q0' if it is
        temperature independent.

        """
        # check the validity of inputs
        if isinstance(shape, str):
            if shape == 'square':
                value = isinstance(initial_point, tuple)
                if value and isinstance(final_point, tuple):
                    cond1 = len(initial_point) == 2
                    cond1 = cond1 and len(final_point) == 2
                else:
                    cond1 = False
            elif shape == 'circle':
                value = isinstance(final_point, int)
                value = value or isinstance(final_point, float)
                value = value and isinstance(initial_point, tuple)
                if value:
                    cond1 = len(initial_point) == 2
                else:
                    cond1 = False
            else:
                cond1 = False
        else:
            cond1 = False
        cond2 = isinstance(power, int) or isinstance(power, float)
        if isinstance(power_type, str):
            if power_type == 'Q' or power_type == 'Q0':
                cond3 = True
            else:
                cond3 = False
        else:
            cond3 = False
        if not cond1 and cond2 and cond3:
            raise ValueError

        if power_type == 'Q':
            if shape == 'square':
                initial_point_x = int(initial_point[0])
                initial_point_y = int(initial_point[1])
                final_point_x = int(final_point[0])
                final_point_y = int(final_point[1])
                for i in range(initial_point_x, final_point_x):
                    for j in range(initial_point_y, final_point_y):
                        self.Q[i][j] = power
            if shape == 'circle':
                initial_point_x = int(initial_point[0])
                initial_point_y = int(initial_point[1])
                radius = int(final_point)
                for i in range(self.size[0]):
                    for j in range(self.size[1]):
                        value = (i-initial_point_x)**2+(j-initial_point_y)**2
                        length = np.sqrt(value)
                        if length <= radius:
                            self.Q[i][j] = power

        if power_type == 'Q0':
            if shape == 'square':
                initial_point_x = int(initial_point[0])
                initial_point_y = int(initial_point[1])
                final_point_x = int(final_point[0])
                final_point_y = int(final_point[1])
                for i in range(initial_point_x, final_point_x):
                    for j in range(initial_point_y, final_point_y):
                        self.Q0[i][j] = power
            if shape == 'circle':
                initial_point_x = int(initial_point[0])
                initial_point_y = int(initial_point[1])
                radius = int(final_point)
                for i in range(self.size[0]):
                    for j in range(self.size[1]):
                        value = (i-initial_point_x)**2+(j-initial_point_y)**2
                        length = np.sqrt(value)
                        if length <= radius:
                            self.Q0[i][j] = power

    def power_reset(self, power_type='Q'):
        """Power reset.

        Resets the power matrix with power_type 'Q' or 'Q0', which corresponds
        to the power temperature dependent and temperature independent,
        respectively.

        """
        # check the validity of inputs
        if isinstance(power_type, str):
            if power_type == 'Q' or power_type == 'Q0':
                condition = True
            else:
                condition = False
        else:
            condition = False
        if not condition:
            raise ValueError

        self.Q = []
        self.Q0 = []
